@@ -18,6 +18,7 @@
 
 #include "commands/explain.h"
 #include "executor/executor.h"
+#include "distributed/citus_nodefuncs.h"
 #include "distributed/commit_protocol.h"
 #include "distributed/master_protocol.h"
 #include "distributed/multi_copy.h"
@@ -29,6 +30,7 @@
 #include "distributed/multi_router_executor.h"
 #include "distributed/multi_router_planner.h"
 #include "distributed/multi_server_executor.h"
+#include "distributed/multi_shard_transaction.h"
 #include "distributed/multi_utility.h"
 #include "distributed/task_tracker.h"
 #include "distributed/worker_manager.h"
@@ -128,6 +130,9 @@ _PG_init(void)
 	 */
 	RegisterCitusConfigVariables();
 
+	/* make our additional node types known */
+	RegisterNodes();
+
 	/* intercept planner */
 	planner_hook = multi_planner;
 
@@ -149,11 +154,9 @@ _PG_init(void)
 	/* organize that task tracker is started once server is up */
 	TaskTrackerRegister();
 
-	/* initialize worker node manager */
-	WorkerNodeRegister();
-
-	/* initialize router executor callbacks */
-	InstallRouterExecutorShmemHook();
+	/* initialize transaction callbacks */
+	RegisterRouterExecutorXactCallbacks();
+	RegisterShardPlacementXactCallbacks();
 }
 
 
@@ -191,6 +194,7 @@ CreateRequiredDirectories(void)
 static void
 RegisterCitusConfigVariables(void)
 {
+	/* keeping temporarily for updates from pre-6.0 versions */
 	DefineCustomStringVariable(
 		"citus.worker_list_file",
 		gettext_noop("Sets the server's \"worker_list\" configuration file."),
@@ -198,7 +202,7 @@ RegisterCitusConfigVariables(void)
 		&WorkerListFileName,
 		NULL,
 		PGC_POSTMASTER,
-		GUC_SUPERUSER_ONLY,
+		GUC_SUPERUSER_ONLY | GUC_NO_SHOW_ALL,
 		NULL, NULL, NULL);
 	NormalizeWorkerListPath();
 
@@ -328,6 +332,17 @@ RegisterCitusConfigVariables(void)
 		0,
 		NULL, NULL, NULL);
 
+	DefineCustomIntVariable(
+		"citus.shard_count",
+		gettext_noop("Sets the number of shards for a new hash-partitioned table"
+					 "created with create_distributed_table()."),
+		NULL,
+		&ShardCount,
+		32, 1, 64000,
+		PGC_USERSET,
+		0,
+		NULL, NULL, NULL);
+
 	DefineCustomIntVariable(
 		"citus.shard_replication_factor",
 		gettext_noop("Sets the replication factor for shards."),
@@ -392,7 +407,7 @@ RegisterCitusConfigVariables(void)
 					 "before walking over these tasks again. This configuration "
 					 "value determines the length of that sleeping period."),
 		&TaskTrackerDelay,
-		200, 10, 100000,
+		200, 1, 100000,
 		PGC_SIGHUP,
 		GUC_UNIT_MS,
 		NULL, NULL, NULL);
