@@ -18,16 +18,22 @@
 #include "access/xact.h"
 #include "access/sysattr.h"
 #include "catalog/indexing.h"
+#include "catalog/pg_am.h"
 #include "catalog/pg_extension.h"
 #include "catalog/pg_namespace.h"
 #include "catalog/pg_type.h"
 #include "commands/extension.h"
 #include "commands/trigger.h"
+#include "distributed/colocation_utils.h"
 #include "distributed/master_metadata_utility.h"
 #include "distributed/metadata_cache.h"
+#include "distributed/pg_dist_local_group.h"
+#include "distributed/pg_dist_node.h"
 #include "distributed/pg_dist_partition.h"
 #include "distributed/pg_dist_shard.h"
+#include "distributed/pg_dist_shard_placement.h"
 #include "distributed/shardinterval_utils.h"
+#include "distributed/worker_manager.h"
 #include "distributed/worker_protocol.h"
 #include "parser/parse_func.h"
 #include "utils/builtins.h"
@@ -48,16 +54,34 @@
 static bool extensionLoaded = false;
 static Oid distShardRelationId = InvalidOid;
 static Oid distShardPlacementRelationId = InvalidOid;
+static Oid distNodeRelationId = InvalidOid;
+static Oid distLocalGroupRelationId = InvalidOid;
+static Oid distColocationRelationId = InvalidOid;
+static Oid distColocationConfigurationIndexId = InvalidOid;
 static Oid distPartitionRelationId = InvalidOid;
 static Oid distPartitionLogicalRelidIndexId = InvalidOid;
+static Oid distPartitionColocationidIndexId = InvalidOid;
 static Oid distShardLogicalRelidIndexId = InvalidOid;
 static Oid distShardShardidIndexId = InvalidOid;
 static Oid distShardPlacementShardidIndexId = InvalidOid;
+static Oid distShardPlacementPlacementidIndexId = InvalidOid;
+static Oid distShardPlacementNodeidIndexId = InvalidOid;
+static Oid distTransactionRelationId = InvalidOid;
+static Oid distTransactionGroupIndexId = InvalidOid;
 static Oid extraDataContainerFuncId = InvalidOid;
 
 /* Hash table for informations about each partition */
 static HTAB *DistTableCacheHash = NULL;
 
+/* Hash table for informations about worker nodes */
+static HTAB *WorkerNodeHash = NULL;
+static bool workerNodeHashValid = false;
+
+static bool invalidationRegistered = false;
+
+/* default value is -1, for schema node it's 0 and for worker nodes > 0 */
+static int LocalGroupId = -1;
+
 /* built first time through in InitializePartitionCache */
 static ScanKeyData DistPartitionScanKey[1];
 static ScanKeyData DistShardScanKey[1];
@@ -76,8 +100,13 @@ static bool HasUniformHashDistribution(ShardInterval **shardIntervalArray,
 static bool HasUninitializedShardInterval(ShardInterval **sortedShardIntervalArray,
 										  int shardCount);
 static void InitializeDistTableCache(void);
+static void InitializeWorkerNodeCache(void);
+static uint32 WorkerNodeHashCode(const void *key, Size keySize);
 static void ResetDistTableCacheEntry(DistTableCacheEntry *cacheEntry);
 static void InvalidateDistRelationCacheCallback(Datum argument, Oid relationId);
+static void InvalidateNodeRelationCacheCallback(Datum argument, Oid relationId);
+static List * DistTableOidList(void);
+static void InvalidateLocalGroupIdRelationCacheCallback(Datum argument, Oid relationId);
 static HeapTuple LookupDistPartitionTuple(Relation pgDistPartition, Oid relationId);
 static List * LookupDistShardTuples(Oid relationId);
 static void GetPartitionTypeInputInfo(char *partitionKeyString, char partitionMethod,
@@ -91,6 +120,9 @@ static void CachedRelationLookup(const char *relationName, Oid *cachedOid);
 /* exports for SQL callable functions */
 PG_FUNCTION_INFO_V1(master_dist_partition_cache_invalidate);
 PG_FUNCTION_INFO_V1(master_dist_shard_cache_invalidate);
+PG_FUNCTION_INFO_V1(master_dist_placement_cache_invalidate);
+PG_FUNCTION_INFO_V1(master_dist_node_cache_invalidate);
+PG_FUNCTION_INFO_V1(master_dist_local_group_cache_invalidate);
 
 
 /*
@@ -118,6 +150,34 @@ IsDistributedTable(Oid relationId)
 }
 
 
+/*
+ * DistributedTableList returns a list that includes all the valid distributed table
+ * cache entries.
+ */
+List *
+DistributedTableList(void)
+{
+	List *distTableOidList = NIL;
+	List *distributedTableList = NIL;
+	ListCell *distTableOidCell = NULL;
+
+	/* first, we need to iterate over pg_dist_partition */
+	distTableOidList = DistTableOidList();
+
+	foreach(distTableOidCell, distTableOidList)
+	{
+		DistTableCacheEntry *cacheEntry = NULL;
+		Oid relationId = lfirst_oid(distTableOidCell);
+
+		cacheEntry = DistributedTableCacheEntry(relationId);
+
+		distributedTableList = lappend(distributedTableList, cacheEntry);
+	}
+
+	return distributedTableList;
+}
+
+
 /*
  * LoadShardInterval reads shard metadata for given shardId from pg_dist_shard,
  * and converts min/max values in these metadata to their properly typed datum
@@ -218,6 +278,8 @@ LookupDistTableCacheEntry(Oid relationId)
 	HeapTuple distPartitionTuple = NULL;
 	char *partitionKeyString = NULL;
 	char partitionMethod = 0;
+	uint32 colocationId = INVALID_COLOCATION_ID;
+	char replicationModel = 0;
 	List *distShardTupleList = NIL;
 	int shardIntervalArrayLength = 0;
 	ShardInterval **shardIntervalArray = NULL;
@@ -255,18 +317,43 @@ LookupDistTableCacheEntry(Oid relationId)
 		Form_pg_dist_partition partitionForm =
 			(Form_pg_dist_partition) GETSTRUCT(distPartitionTuple);
 		Datum partitionKeyDatum = 0;
+		Datum replicationModelDatum = 0;
 		MemoryContext oldContext = NULL;
+		TupleDesc tupleDescriptor = RelationGetDescr(pgDistPartition);
 		bool isNull = false;
 
 		partitionKeyDatum = heap_getattr(distPartitionTuple,
 										 Anum_pg_dist_partition_partkey,
-										 RelationGetDescr(pgDistPartition),
+										 tupleDescriptor,
 										 &isNull);
 		Assert(!isNull);
 
+		colocationId = heap_getattr(distPartitionTuple,
+									Anum_pg_dist_partition_colocationid, tupleDescriptor,
+									&isNull);
+		if (isNull)
+		{
+			colocationId = INVALID_COLOCATION_ID;
+		}
+
+		replicationModelDatum = heap_getattr(distPartitionTuple,
+											 Anum_pg_dist_partition_repmodel,
+											 tupleDescriptor,
+											 &isNull);
+
+		if (isNull)
+		{
+			/*
+			 * repmodel is NOT NULL but before ALTER EXTENSION citus UPGRADE the column
+			 * doesn't exist
+			 */
+			replicationModelDatum = CharGetDatum('c');
+		}
+
 		oldContext = MemoryContextSwitchTo(CacheMemoryContext);
 		partitionKeyString = TextDatumGetCString(partitionKeyDatum);
 		partitionMethod = partitionForm->partmethod;
+		replicationModel = DatumGetChar(replicationModelDatum);
 
 		MemoryContextSwitchTo(oldContext);
 
@@ -378,6 +465,8 @@ LookupDistTableCacheEntry(Oid relationId)
 		cacheEntry->isDistributedTable = true;
 		cacheEntry->partitionKeyString = partitionKeyString;
 		cacheEntry->partitionMethod = partitionMethod;
+		cacheEntry->colocationId = colocationId;
+		cacheEntry->replicationModel = replicationModel;
 		cacheEntry->shardIntervalArrayLength = shardIntervalArrayLength;
 		cacheEntry->sortedShardIntervalArray = sortedShardIntervalArray;
 		cacheEntry->shardIntervalCompareFunction = shardIntervalCompareFunction;
@@ -593,6 +682,47 @@ DistShardPlacementRelationId(void)
 }
 
 
+/* return oid of pg_dist_node relation */
+Oid
+DistNodeRelationId(void)
+{
+	CachedRelationLookup("pg_dist_node", &distNodeRelationId);
+
+	return distNodeRelationId;
+}
+
+
+/* return oid of pg_dist_local_group relation */
+Oid
+DistLocalGroupIdRelationId(void)
+{
+	CachedRelationLookup("pg_dist_local_group", &distLocalGroupRelationId);
+
+	return distLocalGroupRelationId;
+}
+
+
+/* return oid of pg_dist_colocation relation */
+Oid
+DistColocationRelationId(void)
+{
+	CachedRelationLookup("pg_dist_colocation", &distColocationRelationId);
+
+	return distColocationRelationId;
+}
+
+
+/* return oid of pg_dist_colocation_configuration_index index */
+Oid
+DistColocationConfigurationIndexId(void)
+{
+	CachedRelationLookup("pg_dist_colocation_configuration_index",
+						 &distColocationConfigurationIndexId);
+
+	return distColocationConfigurationIndexId;
+}
+
+
 /* return oid of pg_dist_partition relation */
 Oid
 DistPartitionRelationId(void)
@@ -614,6 +744,17 @@ DistPartitionLogicalRelidIndexId(void)
 }
 
 
+/* return oid of pg_dist_partition_colocationid_index index */
+Oid
+DistPartitionColocationidIndexId(void)
+{
+	CachedRelationLookup("pg_dist_partition_colocationid_index",
+						 &distPartitionColocationidIndexId);
+
+	return distPartitionColocationidIndexId;
+}
+
+
 /* return oid of pg_dist_shard_logical_relid_index index */
 Oid
 DistShardLogicalRelidIndexId(void)
@@ -646,6 +787,49 @@ DistShardPlacementShardidIndexId(void)
 }
 
 
+/* return oid of pg_dist_shard_placement_shardid_index */
+Oid
+DistShardPlacementPlacementidIndexId(void)
+{
+	CachedRelationLookup("pg_dist_shard_placement_placementid_index",
+						 &distShardPlacementPlacementidIndexId);
+
+	return distShardPlacementPlacementidIndexId;
+}
+
+
+/* return oid of pg_dist_transaction relation */
+Oid
+DistTransactionRelationId(void)
+{
+	CachedRelationLookup("pg_dist_transaction", &distTransactionRelationId);
+
+	return distTransactionRelationId;
+}
+
+
+/* return oid of pg_dist_transaction_group_index */
+Oid
+DistTransactionGroupIndexId(void)
+{
+	CachedRelationLookup("pg_dist_transaction_group_index",
+						 &distTransactionGroupIndexId);
+
+	return distTransactionGroupIndexId;
+}
+
+
+/* return oid of pg_dist_shard_placement_nodeid_index */
+Oid
+DistShardPlacementNodeidIndexId(void)
+{
+	CachedRelationLookup("pg_dist_shard_placement_nodeid_index",
+						 &distShardPlacementNodeidIndexId);
+
+	return distShardPlacementNodeidIndexId;
+}
+
+
 /* return oid of the citus_extradata_container(internal) function */
 Oid
 CitusExtraDataContainerFuncId(void)
@@ -730,6 +914,18 @@ CitusExtensionOwner(void)
 }
 
 
+/*
+ * CitusExtensionOwnerName returns the name of the owner of the extension.
+ */
+char *
+CitusExtensionOwnerName(void)
+{
+	Oid superUserId = CitusExtensionOwner();
+
+	return GetUserNameFromId(superUserId, false);
+}
+
+
 /* return the  username of the currently active role */
 char *
 CurrentUserName(void)
@@ -862,6 +1058,114 @@ master_dist_shard_cache_invalidate(PG_FUNCTION_ARGS)
 }
 
 
+/*
+ * master_dist_placmeent_cache_invalidate is a trigger function that performs
+ * relcache invalidations when the contents of pg_dist_shard_placement are
+ * changed on the SQL level.
+ *
+ * NB: We decided there is little point in checking permissions here, there
+ * are much easier ways to waste CPU than causing cache invalidations.
+ */
+Datum
+master_dist_placement_cache_invalidate(PG_FUNCTION_ARGS)
+{
+	TriggerData *triggerData = (TriggerData *) fcinfo->context;
+	HeapTuple newTuple = NULL;
+	HeapTuple oldTuple = NULL;
+	Oid oldShardId = InvalidOid;
+	Oid newShardId = InvalidOid;
+
+	if (!CALLED_AS_TRIGGER(fcinfo))
+	{
+		ereport(ERROR, (errcode(ERRCODE_E_R_I_E_TRIGGER_PROTOCOL_VIOLATED),
+						errmsg("must be called as trigger")));
+	}
+
+	newTuple = triggerData->tg_newtuple;
+	oldTuple = triggerData->tg_trigtuple;
+
+	/* collect shardid for OLD and NEW tuple */
+	if (oldTuple != NULL)
+	{
+		Form_pg_dist_shard_placement distPlacement =
+			(Form_pg_dist_shard_placement) GETSTRUCT(oldTuple);
+
+		oldShardId = distPlacement->shardid;
+	}
+
+	if (newTuple != NULL)
+	{
+		Form_pg_dist_shard_placement distPlacement =
+			(Form_pg_dist_shard_placement) GETSTRUCT(newTuple);
+
+		newShardId = distPlacement->shardid;
+	}
+
+	/*
+	 * Invalidate relcache for the relevant relation(s). In theory shardId
+	 * should never change, but it doesn't hurt to be paranoid.
+	 */
+	if (oldShardId != InvalidOid &&
+		oldShardId != newShardId)
+	{
+		CitusInvalidateRelcacheByShardId(oldShardId);
+	}
+
+	if (newShardId != InvalidOid)
+	{
+		CitusInvalidateRelcacheByShardId(newShardId);
+	}
+
+	PG_RETURN_DATUM(PointerGetDatum(NULL));
+}
+
+
+/*
+ * master_dist_node_cache_invalidate is a trigger function that performs
+ * relcache invalidations when the contents of pg_dist_node are changed
+ * on the SQL level.
+ *
+ * NB: We decided there is little point in checking permissions here, there
+ * are much easier ways to waste CPU than causing cache invalidations.
+ */
+Datum
+master_dist_node_cache_invalidate(PG_FUNCTION_ARGS)
+{
+	if (!CALLED_AS_TRIGGER(fcinfo))
+	{
+		ereport(ERROR, (errcode(ERRCODE_E_R_I_E_TRIGGER_PROTOCOL_VIOLATED),
+						errmsg("must be called as trigger")));
+	}
+
+	CitusInvalidateRelcacheByRelid(DistNodeRelationId());
+
+	PG_RETURN_DATUM(PointerGetDatum(NULL));
+}
+
+
+/*
+ * master_dist_local_group_cache_invalidate is a trigger function that performs
+ * relcache invalidations when the contents of pg_dist_local_group are changed
+ * on the SQL level.
+ *
+ * NB: We decided there is little point in checking permissions here, there
+ * are much easier ways to waste CPU than causing cache invalidations.
+ */
+Datum
+master_dist_local_group_cache_invalidate(PG_FUNCTION_ARGS)
+{
+	if (!CALLED_AS_TRIGGER(fcinfo))
+	{
+		ereport(ERROR, (errcode(ERRCODE_E_R_I_E_TRIGGER_PROTOCOL_VIOLATED),
+						errmsg("must be called as trigger")));
+	}
+
+	CitusInvalidateRelcacheByRelid(DistLocalGroupIdRelationId());
+
+	PG_RETURN_DATUM(PointerGetDatum(NULL));
+}
+
+
 /* initialize the infrastructure for the metadata cache */
 static void
 InitializeDistTableCache(void)
@@ -910,6 +1214,203 @@ InitializeDistTableCache(void)
 }
 
 
+/*
+ * GetWorkerNodeHash is a wrapper around InitializeWorkerNodeCache(). It
+ * triggers InitializeWorkerNodeCache when the workerHash is invalid. Otherwise,
+ * it returns the hash.
+ */
+HTAB *
+GetWorkerNodeHash(void)
+{
+	if (!workerNodeHashValid)
+	{
+		InitializeWorkerNodeCache();
+
+		workerNodeHashValid = true;
+	}
+
+	return WorkerNodeHash;
+}
+
+
+/*
+ * InitializeWorkerNodeCache initialize the infrastructure for the worker node cache.
+ * The function reads the worker nodes from the metadata table, adds them to the hash and
+ * finally registers an invalidation callback.
+ */
+static void
+InitializeWorkerNodeCache(void)
+{
+	static bool invalidationRegistered = false;
+	HTAB *oldWorkerNodeHash = NULL;
+	List *workerNodeList = NIL;
+	ListCell *workerNodeCell = NULL;
+	HASHCTL info;
+	int hashFlags = 0;
+	long maxTableSize = (long) MaxWorkerNodesTracked;
+
+	/* make sure we've initialized CacheMemoryContext */
+	if (CacheMemoryContext == NULL)
+	{
+		CreateCacheMemoryContext();
+	}
+
+	/*
+	 * Create the hash that holds the worker nodes. The key is the combination of
+	 * nodename and nodeport, instead of the unique nodeid because worker nodes are
+	 * searched by the nodename and nodeport in every physical plan creation.
+	 */
+	memset(&info, 0, sizeof(info));
+	info.keysize = +sizeof(uint32) + WORKER_LENGTH + sizeof(uint32);
+	info.entrysize = sizeof(WorkerNode);
+	info.hcxt = CacheMemoryContext;
+	info.hash = WorkerNodeHashCode;
+	info.match = WorkerNodeCompare;
+	hashFlags = HASH_ELEM | HASH_FUNCTION | HASH_CONTEXT | HASH_COMPARE;
+
+	oldWorkerNodeHash = WorkerNodeHash;
+	WorkerNodeHash = hash_create("Worker Node Hash",
+								 maxTableSize,
+								 &info, hashFlags);
+
+	/* read the list from pg_dist_node */
+	workerNodeList = ReadWorkerNodes();
+
+	/* iterate over the worker node list */
+	foreach(workerNodeCell, workerNodeList)
+	{
+		WorkerNode *workerNode = NULL;
+		WorkerNode *currentNode = lfirst(workerNodeCell);
+		void *hashKey = NULL;
+		bool handleFound = false;
+
+		/* search for the worker node in the hash, and then insert the values */
+		hashKey = (void *) currentNode;
+		workerNode = (WorkerNode *) hash_search(WorkerNodeHash, hashKey,
+												HASH_ENTER, &handleFound);
+
+		/* fill the newly allocated workerNode in the cache */
+		strlcpy(workerNode->workerName, currentNode->workerName, WORKER_LENGTH);
+		workerNode->workerPort = currentNode->workerPort;
+		workerNode->groupId = currentNode->groupId;
+		workerNode->nodeId = currentNode->nodeId;
+		strlcpy(workerNode->workerRack, currentNode->workerRack, WORKER_LENGTH);
+		workerNode->hasMetadata = currentNode->hasMetadata;
+
+		if (handleFound)
+		{
+			ereport(WARNING, (errmsg("multiple lines for worker node: \"%s:%u\"",
+									 workerNode->workerName,
+									 workerNode->workerPort)));
+		}
+
+		/* we do not need the currentNode anymore */
+		pfree(currentNode);
+	}
+
+	/* now, safe to destroy the old hash */
+	hash_destroy(oldWorkerNodeHash);
+
+	/* prevent multiple invalidation registrations */
+	if (!invalidationRegistered)
+	{
+		/* Watch for invalidation events. */
+		CacheRegisterRelcacheCallback(InvalidateNodeRelationCacheCallback,
+									  (Datum) 0);
+
+		invalidationRegistered = true;
+	}
+}
+
+
+/*
+ * GetLocalGroupId returns the group identifier of the local node. The function assumes
+ * that pg_dist_local_node_group has exactly one row and has at least one column.
+ * Otherwise, the function errors out.
+ */
+int
+GetLocalGroupId(void)
+{
+	SysScanDesc scanDescriptor = NULL;
+	ScanKeyData scanKey[1];
+	int scanKeyCount = 0;
+	HeapTuple heapTuple = NULL;
+	TupleDesc tupleDescriptor = NULL;
+	Oid groupId = InvalidOid;
+	Relation pgDistLocalGroupId = NULL;
+
+	/*
+	 * Already set the group id, no need to read the heap again.
+	 */
+	if (LocalGroupId != -1)
+	{
+		return LocalGroupId;
+	}
+
+	pgDistLocalGroupId = heap_open(DistLocalGroupIdRelationId(), AccessShareLock);
+
+	scanDescriptor = systable_beginscan(pgDistLocalGroupId,
+										InvalidOid, false,
+										NULL, scanKeyCount, scanKey);
+
+	tupleDescriptor = RelationGetDescr(pgDistLocalGroupId);
+
+	heapTuple = systable_getnext(scanDescriptor);
+
+	if (HeapTupleIsValid(heapTuple))
+	{
+		bool isNull = false;
+		Datum groupIdDatum = heap_getattr(heapTuple,
+										  Anum_pg_dist_local_groupid,
+										  tupleDescriptor, &isNull);
+
+		groupId = DatumGetUInt32(groupIdDatum);
+	}
+	else
+	{
+		elog(ERROR, "could not find any entries in pg_dist_local_group");
+	}
+
+	systable_endscan(scanDescriptor);
+	heap_close(pgDistLocalGroupId, AccessShareLock);
+
+	/* prevent multiple invalidation registrations */
+	if (!invalidationRegistered)
+	{
+		/* Watch for invalidation events. */
+		CacheRegisterRelcacheCallback(InvalidateLocalGroupIdRelationCacheCallback,
+									  (Datum) 0);
+
+		invalidationRegistered = true;
+	}
+
+	/* set the local cache variable */
+	LocalGroupId = groupId;
+
+	return groupId;
+}
+
+
+/*
+ * WorkerNodeHashCode computes the hash code for a worker node from the node's
+ * host name and port number. Nodes that only differ by their rack locations
+ * hash to the same value.
+ */
+static uint32
+WorkerNodeHashCode(const void *key, Size keySize)
+{
+	const WorkerNode *worker = (const WorkerNode *) key;
+	const char *workerName = worker->workerName;
+	const uint32 *workerPort = &(worker->workerPort);
+
+	/* standard hash function outlined in Effective Java, Item 8 */
+	uint32 result = 17;
+	result = 37 * result + string_hash(workerName, WORKER_LENGTH);
+	result = 37 * result + tag_hash(workerPort, sizeof(uint32));
+	return result;
+}
+
+
 /*
  * ResetDistTableCacheEntry frees any out-of-band memory used by a cache entry,
  * but does not free the entry itself.
@@ -1011,16 +1512,100 @@ InvalidateDistRelationCacheCallback(Datum argument, Oid relationId)
 		extensionLoaded = false;
 		distShardRelationId = InvalidOid;
 		distShardPlacementRelationId = InvalidOid;
+		distLocalGroupRelationId = InvalidOid;
+		distNodeRelationId = InvalidOid;
+		distColocationRelationId = InvalidOid;
+		distColocationConfigurationIndexId = InvalidOid;
 		distPartitionRelationId = InvalidOid;
 		distPartitionLogicalRelidIndexId = InvalidOid;
+		distPartitionColocationidIndexId = InvalidOid;
 		distShardLogicalRelidIndexId = InvalidOid;
 		distShardShardidIndexId = InvalidOid;
 		distShardPlacementShardidIndexId = InvalidOid;
+		distShardPlacementPlacementidIndexId = InvalidOid;
+		distTransactionRelationId = InvalidOid;
+		distTransactionGroupIndexId = InvalidOid;
 		extraDataContainerFuncId = InvalidOid;
 	}
 }
 
 
+/*
+ * DistTableOidList iterates over the pg_dist_partition table and returns
+ * a list that consists of the logicalrelids.
+ */
+static List *
+DistTableOidList(void)
+{
+	SysScanDesc scanDescriptor = NULL;
+	ScanKeyData scanKey[1];
+	int scanKeyCount = 0;
+	HeapTuple heapTuple = NULL;
+	List *distTableOidList = NIL;
+	TupleDesc tupleDescriptor = NULL;
+
+	Relation pgDistPartition = heap_open(DistPartitionRelationId(), AccessShareLock);
+
+	scanDescriptor = systable_beginscan(pgDistPartition,
+										InvalidOid, false,
+										NULL, scanKeyCount, scanKey);
+
+	tupleDescriptor = RelationGetDescr(pgDistPartition);
+
+	heapTuple = systable_getnext(scanDescriptor);
+	while (HeapTupleIsValid(heapTuple))
+	{
+		bool isNull = false;
+		Oid relationId = InvalidOid;
+		Datum relationIdDatum = heap_getattr(heapTuple,
+											 Anum_pg_dist_partition_logicalrelid,
+											 tupleDescriptor, &isNull);
+
+		relationId = DatumGetObjectId(relationIdDatum);
+
+		distTableOidList = lappend_oid(distTableOidList, relationId);
+
+		heapTuple = systable_getnext(scanDescriptor);
+	}
+
+	systable_endscan(scanDescriptor);
+	heap_close(pgDistPartition, AccessShareLock);
+
+	return distTableOidList;
+}
+
+
+/*
+ * InvalidateNodeRelationCacheCallback destroys the WorkerNodeHash when
+ * any change happens on pg_dist_node table. It also set WorkerNodeHash to
+ * NULL, which allows consequent accesses to the hash read from the
+ * pg_dist_node from scratch.
+ */
+static void
+InvalidateNodeRelationCacheCallback(Datum argument, Oid relationId)
+{
+	if (relationId == InvalidOid || relationId == distNodeRelationId)
+	{
+		workerNodeHashValid = false;
+	}
+}
+
+
+/*
+ * InvalidateLocalGroupIdRelationCacheCallback sets the LocalGroupId to
+ * the default value.
+ */
+static void
+InvalidateLocalGroupIdRelationCacheCallback(Datum argument, Oid relationId)
+{
+	/* when invalidation happens simply set the LocalGroupId to the default value */
+	if (relationId == InvalidOid || relationId == distLocalGroupRelationId)
+	{
+		LocalGroupId = -1;
+	}
+}
+
+
 /*
  * LookupDistPartitionTuple searches pg_dist_partition for relationId's entry
  * and returns that or, if no matching entry was found, NULL.
@@ -1251,3 +1836,64 @@ CitusInvalidateRelcacheByRelid(Oid relationId)
 		ReleaseSysCache(classTuple);
 	}
 }
+
+
+/*
+ * Register a relcache invalidation for the distributed relation associated
+ * with the shard.
+ */
+void
+CitusInvalidateRelcacheByShardId(int64 shardId)
+{
+	SysScanDesc scanDescriptor = NULL;
+	ScanKeyData scanKey[1];
+	int scanKeyCount = 1;
+	HeapTuple heapTuple = NULL;
+	Form_pg_dist_shard shardForm = NULL;
+	Relation pgDistShard = heap_open(DistShardRelationId(), AccessShareLock);
+
+	/*
+	 * Load shard, to find the associated relation id. Can't use
+	 * LoadShardInterval directly because that'd fail if the shard doesn't
+	 * exist anymore, which we can't have. Also lower overhead is desirable
+	 * here.
+	 */
+
+	ScanKeyInit(&scanKey[0], Anum_pg_dist_shard_shardid,
+				BTEqualStrategyNumber, F_INT8EQ, Int64GetDatum(shardId));
+
+	scanDescriptor = systable_beginscan(pgDistShard,
+										DistShardShardidIndexId(), true,
+										NULL, scanKeyCount, scanKey);
+
+	heapTuple = systable_getnext(scanDescriptor);
+	if (HeapTupleIsValid(heapTuple))
+	{
+		shardForm = (Form_pg_dist_shard) GETSTRUCT(heapTuple);
+		CitusInvalidateRelcacheByRelid(shardForm->logicalrelid);
+	}
+	else
+	{
+		/*
+		 * Couldn't find associated relation. That can primarily happen in two cases:
+		 *
+		 * 1) A placement row is inserted before the shard row. That's fine,
+		 *	  since we don't need invalidations via placements in that case.
+		 *
+		 * 2) The shard has been deleted, but some placements were
+		 *    unreachable, and the user is manually deleting the rows. Not
+		 *    much point in WARNING or ERRORing in that case either, there's
+		 *    nothing to invalidate.
+		 *
+		 * Hence we just emit a DEBUG5 message.
+		 */
+		ereport(DEBUG5, (errmsg("could not find distributed relation to invalidate for "
+								"shard "INT64_FORMAT, shardId)));
+	}
+
+	systable_endscan(scanDescriptor);
+	heap_close(pgDistShard, NoLock);
+
+	/* bump command counter, to force invalidation to take effect */
+	CommandCounterIncrement();
+}
